--[[
Penalty camera module, written by nesa24, Chuny, and digitalfoxx
modified by juce to be exe-adaptable.
--]]

local m = {}
m.version = "1.2"
local hex = memory.hex
local settings
 
local RESTORE_KEY = 0x38
local PREV_PROP_KEY = 0x39
local NEXT_PROP_KEY = 0x30
local PREV_VALUE_KEY = 0xbd
local NEXT_VALUE_KEY = 0xbb
 
local delta = 0
local frame_count = 0
 
local overlay_curr = 1
 
local overlay_states = {
    { ui = "Penalty Height: %0.2f", prop = "Penalty_height", decr = -0.2, incr = 0.2 },
    { ui = "Penalty Zoom: %0.2f", prop = "Penalty_zoom", decr = -0.3, incr = 0.3 },
  
}

local ui_lines = {}
 
local bases = {}
local game_info = {
    Penalty_zoom = { base = "pzoom", offs = 0x00, format = "f", len = 4, def = 17.60000038 },
    Penalty_height   = { base = "pheight", offs = 0x00, format = "f", len = 4, def = 3 },
}
 
local function save_ini(ctx, filename)
    local f,err = io.open(ctx.sider_dir .. "\\" .. filename, "wt")
    if not f then
        log(string.format("PROBLEM saving settings: %s", tostring(err)))
        return
    end
    f:write(string.format("# Penalty Zoom_Height settings by PESWEB \n\n"))
    f:write(string.format("Penalty_zoom = %0.2f\n", settings.Penalty_zoom or game_info.Penalty_zoom.def))
    f:write(string.format("Penalty_height = %0.2f\n", settings.Penalty_height or game_info.Penalty_height.def))
    f:write("\n")
    f:close()
end
 
local function load_ini(ctx, filename)
    local t = {}
    -- initialize with defaults
    for prop,info in pairs(game_info) do
        t[prop] = info.def
    end
    -- now try to read ini-file, if present
    local f,err = io.open(ctx.sider_dir .. "\\" .. filename)
    if not f then
        return t
    end
    f:close()
    for line in io.lines(ctx.sider_dir .. "\\" .. filename) do
        local name, value = string.match(line, "^([%w_]+)%s*=%s*([-%w%d.]+)")
        if name and value then
            value = tonumber(value) or value
            t[name] = value
            log(string.format("Using setting: %s = %s", name, value))
        end
    end
    return t
end
 
local function apply_settings(ctx, log_it, save_it)
    for name,value in pairs(settings) do
        local entry = game_info[name]
        if entry then
            local base = bases[entry.base]
            if base then
                if entry.value_map then
                    value = entry.value_map[value]
                end
                local addr = base + entry.offs
                local old_value, new_value
                if entry.format ~= "" then
                    old_value = memory.unpack(entry.format, memory.read(addr, entry.len))
                    memory.write(addr, memory.pack(entry.format, value))
                    new_value = memory.unpack(entry.format, memory.read(addr, entry.len))
                    if log_it then
                        log(string.format("%s: changed at %s: %s --> %s",
                            name, hex(addr), old_value, new_value))
                    end
                else
                    old_value = memory.read(addr, entry.len)
                    memory.write(addr, value)
                    new_value = memory.read(addr, entry.len)
                    if log_it then
                        log(string.format("%s: changed at %s: %s --> %s",
                            name, hex(addr), hex(old_value), hex(new_value)))
                    end
                end
            end
        end
    end
    if (save_it) then
        save_ini(ctx, "modules\\Penalty_zh.ini")
    end
end
 
local function repeat_change(ctx, after_num_frames, change)
    if change ~= 0 then
        frame_count = frame_count + 1
        if frame_count >= after_num_frames then
            local s = overlay_states[overlay_curr]
            settings[s.prop] = settings[s.prop] + change
            apply_settings(ctx, false, false) -- apply
        end
    end
end
 
function m.overlay_on(ctx)
    -- repeat change from gamepad, if delta exists
    repeat_change(ctx, 30, delta)
    -- construct ui text
    for i,v in ipairs(overlay_states) do
        local s = overlay_states[i]
        local setting = string.format(s.ui, settings[s.prop])
        if i == overlay_curr then
            ui_lines[i] = string.format("\n---> %s <---", setting)
        else
            ui_lines[i] = string.format("\n     %s", setting)
        end
    end
    return string.format([[version %s
Keys: [9][0] - choose setting, [-][+] - modify value, [8] - restore defaults
Gamepad: RS up/down - choose setting, RS left/right - modify value
%s]], m.version, table.concat(ui_lines))
end
 
function m.key_down(ctx, vkey)
    if vkey == NEXT_PROP_KEY then
        if overlay_curr < #overlay_states then
            overlay_curr = overlay_curr + 1
        end
    elseif vkey == PREV_PROP_KEY then
        if overlay_curr > 1 then
            overlay_curr = overlay_curr - 1
        end
    elseif vkey == NEXT_VALUE_KEY then
        local s = overlay_states[overlay_curr]
        if s.incr ~= nil then
            settings[s.prop] = settings[s.prop] + s.incr
        elseif s.nextf ~= nil then
            settings[s.prop] = s.nextf(settings[s.prop])
        end
        apply_settings(ctx, false, true)
    elseif vkey == PREV_VALUE_KEY then
        local s = overlay_states[overlay_curr]
        if s.decr ~= nil then
            settings[s.prop] = settings[s.prop] + s.decr
        elseif s.prevf ~= nil then
            settings[s.prop] = s.prevf(settings[s.prop])
        end
        apply_settings(ctx, false, true)
    elseif vkey == RESTORE_KEY then
        for i,s in ipairs(overlay_states) do
            settings[s.prop] = game_info[s.prop].def
        end
        apply_settings(ctx, false, true)
    end
end
 
function m.gamepad_input(ctx, inputs)
    local v = inputs["RSy"]
    if v then
        if v == -1 and overlay_curr < #overlay_states then -- moving down
            overlay_curr = overlay_curr + 1
        elseif v == 1 and overlay_curr > 1 then -- moving up
            overlay_curr = overlay_curr - 1
        end
    end
 
    v = inputs["RSx"]
    if v then
        if v == -1 then -- moving left
            local s = overlay_states[overlay_curr]
            if s.decr ~= nil then
                settings[s.prop] = settings[s.prop] + s.decr
                -- set up the repeat change
                delta = s.decr
                frame_count = 0
            elseif s.prevf ~= nil then
                settings[s.prop] = s.prevf(settings[s.prop])
            end
            apply_settings(ctx, false, false) -- apply
        elseif v == 1 then -- moving right
            local s = overlay_states[overlay_curr]
            if s.decr ~= nil then
                settings[s.prop] = settings[s.prop] + s.incr
                -- set up the repeat change
                delta = s.incr
                frame_count = 0
            elseif s.nextf ~= nil then
                settings[s.prop] = s.nextf(settings[s.prop])
            end
            apply_settings(ctx, false, false) -- apply
        elseif v == 0 then -- stop change
            delta = 0
            apply_settings(ctx, false, true) -- apply and save
        end
    end
end

local function find_pattern(ctx, pattern, cache_file, cache_id)
    if cache_id then
        -- check the cached location first: it might match
        local f = io.open(cache_file, "rb")
        if f then
            local cache_data = f:read("*all")
            local hint = string.sub(cache_data, cache_id*8+1, cache_id*8+8)
            if hint and hint ~= "" then
                local addr = memory.unpack("i64", hint)
                if addr and addr ~= 0 then
                    local data = memory.read(addr, #pattern)
                    if pattern == data then
                        log(string.format("matched cache hint #%s", cache_id))
                        return addr
                    end
                end
            end
        end
    end
    -- no cache, or no match: search the process
    return memory.search_process(pattern)
end

local function write_cache(ctx, cache, cache_file, cache_size)
    local f = io.open(cache_file, "wb")
    for i=0,cache_size-1 do
        addr = cache[i] or 0
        f:write(memory.pack("i64", addr))
    end
    f:close()
end

function m.init(ctx)
    settings = load_ini(ctx, "modules\\Penalty_zh.ini")

    local cache = {}
    local cache_file = ctx.sider_dir .. _FILE:gsub("%.lua", ".cache")

    -- 0000000146D634B3 | F3 0F 10 8B 88 08 00 00            | movss xmm1,dword ptr ds:[rbx+888]      |
    -- 0000000146D634BB | F3 0F 59 35 19 53 82 FB            | mulss xmm6,dword ptr ds:[1425887DC]    | 
    local loc = find_pattern(ctx, "\xf3\x0f\x10\x8b\x88\x08\x00\x00\xf3\x0f\x59\x35", cache_file, 0)
    if loc then
        cache[0] = loc
        local rel_offset = memory.unpack("i32", memory.read(loc + 12, 4))
        bases.pzoom = loc + 16 + rel_offset
        log(string.format("penalty camera zoom address: %s", hex(bases.pzoom)))
    else
        error("problem: unable to find code pattern for penalty camera zoom")
    end

    -- 0000000146D65042 | 48 8D 4E 08                        | lea rcx,qword ptr ds:[rsi+8]           | rcx:"p=&B"
    -- 0000000146D65046 | 48 89 9C 24 00 01 00 00            | mov qword ptr ss:[rsp+100],rbx         |
    -- 0000000146D6504E | F3 44 0F 58 35 8D 11 7C FB         | addss xmm14,dword ptr ds:[1425261E4]   | 
    local loc = find_pattern(ctx, "\x48\x8d\x4e\x08\x48\x89\x9c\x24\x00\x01\x00\x00\xf3\x44\x0f\x58\x35", cache_file, 1)
    if loc then
        cache[1] = loc
        local rel_offset = memory.unpack("i32", memory.read(loc + 17, 4))
        -- need to "cave" the value, meaning: change offset to another location in memory
        -- then put the value in there.
        rel_offset = rel_offset + 0x18  -- there is a free/unused 4-bytes
        memory.write(loc + 17, memory.pack("i32", rel_offset))
        bases.pheight = loc + 21 + rel_offset
        log(string.format("penalty camera height address: %s", hex(bases.pheight)))
    else
        error("problem: unable to find code pattern for penalty camera height")
    end

    -- save found locations to disk
    write_cache(ctx, cache, cache_file, 2)

    -- set our new values
    apply_settings(ctx, true)
 
    -- register for events
    -- ctx.register("set_teams", m.set_teams)
    ctx.register("overlay_on", m.overlay_on)
    ctx.register("key_down", m.key_down)
    ctx.register("gamepad_input", m.gamepad_input)
end
 
return m
